@@ -1,7 +1,7 @@
 /*
  * %CopyrightBegin%
  * 
- * Copyright Ericsson AB 2006-2013. All Rights Reserved.
+ * Copyright Ericsson AB 2006-2018. All Rights Reserved.
  * 
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -45,14 +45,6 @@
 #endif
 #endif
 
-/*
- * erts_check_io_time is used by the erl_check_io implementation. The
- * global erts_check_io_time variable is declared here since there
- * (often) exist two versions of erl_check_io (kernel-poll and
- * non-kernel-poll), and we dont want two versions of this variable.
- */
-erts_smp_atomic_t erts_check_io_time;
-
 /* Written once and only once */
 
 static int filename_encoding = ERL_FILENAME_UNKNOWN;
@@ -150,7 +142,16 @@ sys_double_to_chars(double fp, char *buffer, size_t buffer_size)
     return sys_double_to_chars_ext(fp, buffer, buffer_size, SYS_DEFAULT_FLOAT_DECIMALS);
 }
 
-/* Convert float to string using fixed point notation.
+
+#if SIZEOF_LONG == 8
+# define round_int64 lround
+#elif SIZEOF_LONG_LONG == 8
+# define round_int64 llround
+#else
+# error "No 64-bit integer type?"
+#endif
+
+/* Convert float to string
  *   decimals must be >= 0
  *   if compact != 0, the trailing 0's will be truncated
  */
@@ -158,93 +159,39 @@ int
 sys_double_to_chars_fast(double f, char *buffer, int buffer_size, int decimals,
 			 int compact)
 {
-    /* Note that some C compilers don't support "static const" propagation
-     * so we use a defines */
-    #define SYS_DOUBLE_RND_CONST 0.55555555555555555
+    #define SYS_DOUBLE_RND_CONST 0.5
     #define FRAC_SIZE            52
     #define EXP_SIZE             11
-    #define EXP_MASK             ((1ll << EXP_SIZE) - 1)
-    #define MAX_DECIMALS         (sizeof(cs_sys_double_pow10) \
-				   / sizeof(cs_sys_double_pow10[0]))
-    #define FRAC_MASK            ((1ll << FRAC_SIZE) - 1)
-    #define FRAC_MASK2           ((1ll << (FRAC_SIZE + 1)) - 1)
-    #define MAX_FLOAT            (1ll << (FRAC_SIZE+1))
+    #define EXP_MASK             (((Uint64)1 << EXP_SIZE) - 1)
+    #define MAX_DECIMALS         (sizeof(pow10v) / sizeof(pow10v[0]))
+    #define FRAC_MASK            (((Uint64)1 << FRAC_SIZE) - 1)
+    #define FRAC_MASK2           (((Uint64)1 << (FRAC_SIZE + 1)) - 1)
+    #define MAX_FLOAT            ((Uint64)1 << (FRAC_SIZE+1))
 
-    static const double cs_sys_double_pow10[] = {
-        SYS_DOUBLE_RND_CONST / 1ll,
-        SYS_DOUBLE_RND_CONST / 10ll,
-        SYS_DOUBLE_RND_CONST / 100ll,
-        SYS_DOUBLE_RND_CONST / 1000ll,
-        SYS_DOUBLE_RND_CONST / 10000ll,
-        SYS_DOUBLE_RND_CONST / 100000ll,
-        SYS_DOUBLE_RND_CONST / 1000000ll,
-        SYS_DOUBLE_RND_CONST / 10000000ll,
-        SYS_DOUBLE_RND_CONST / 100000000ll,
-        SYS_DOUBLE_RND_CONST / 1000000000ll,
-        SYS_DOUBLE_RND_CONST / 10000000000ll,
-        SYS_DOUBLE_RND_CONST / 100000000000ll,
-        SYS_DOUBLE_RND_CONST / 1000000000000ll,
-        SYS_DOUBLE_RND_CONST / 10000000000000ll,
-        SYS_DOUBLE_RND_CONST / 100000000000000ll,
-        SYS_DOUBLE_RND_CONST / 1000000000000000ll,
-        SYS_DOUBLE_RND_CONST / 10000000000000000ll,
-        SYS_DOUBLE_RND_CONST / 100000000000000000ll,
-        SYS_DOUBLE_RND_CONST / 1000000000000000000ll
+    static const double pow10v[] = {
+        1e0,  1e1,  1e2,  1e3,  1e4,  1e5,  1e6,  1e7,  1e8, 1e9,
+        1e10, 1e11, 1e12, 1e13, 1e14, 1e15, 1e16, 1e17, 1e18
     };
 
-    long long mantissa, int_part = 0, frac_part = 0;
-    short exp;
-    int max;
+    double af;
+    Uint64 int_part, frac_part;
     int neg;
-    double fr;
-    union { long long L; double F; } x;
     char *p = buffer;
 
     if (decimals < 0)
         return -1;
 
-    /* Round the number to given decimal places. The number of 5's in the
-     * SYS_DOUBLE_RND_CONST constant is chosen such that adding any more 5's doesn't
-     * change the double precision of the number, i.e.:
-     * 1> term_to_binary(0.55555555555555555, [{minor_version, 1}]).
-     * <<131,70,63,225,199,28,113,199,28,114>>
-     * 2> term_to_binary(0.5555555555555555555, [{minor_version, 1}]).
-     * <<131,70,63,225,199,28,113,199,28,114>>
-     */
-    if (f >= 0) {
-        neg = 0;
-        fr  = decimals < MAX_DECIMALS ? (f + cs_sys_double_pow10[decimals]) : f;
-        x.F = fr;
-    } else {
+    if (f < 0) {
         neg = 1;
-        fr  = decimals < MAX_DECIMALS ? (f - cs_sys_double_pow10[decimals]) : f;
-        x.F = -fr;
+        af = -f;
     }
-
-    exp      = (x.L >> FRAC_SIZE) & EXP_MASK;
-    mantissa = x.L & FRAC_MASK;
-
-    if (exp == EXP_MASK) {
-        if (mantissa == 0) {
-            if (neg)
-                *p++ = '-';
-            *p++ = 'i';
-            *p++ = 'n';
-            *p++ = 'f';
-        } else {
-            *p++ = 'n';
-            *p++ = 'a';
-            *p++ = 'n';
-        }
-        *p = '\0';
-        return p - buffer;
+    else {
+        neg = 0;
+        af = f;
     }
 
-    exp      -= EXP_MASK >> 1;
-    mantissa |= (1ll << FRAC_SIZE);
-
     /* Don't bother with optimizing too large numbers or too large precision */
-    if (x.F > MAX_FLOAT || decimals >= MAX_DECIMALS) {
+    if (af > MAX_FLOAT || decimals >= MAX_DECIMALS) {
         int len = erts_snprintf(buffer, buffer_size, "%.*f", decimals, f);
         char* p = buffer + len;
         if (len >= buffer_size)
@@ -254,62 +201,64 @@ sys_double_to_chars_fast(double f, char *buffer, int buffer_size, int decimals,
             p = find_first_trailing_zero(p);
         *p = '\0';
         return p - buffer;
-    } else if (exp >= FRAC_SIZE) {
-        int_part  = mantissa << (exp - FRAC_SIZE);
-    } else if (exp >= 0) {
-        int_part  = mantissa >> (FRAC_SIZE - exp);
-        frac_part = (mantissa << (exp + 1)) & FRAC_MASK2;
-    } else /* if (exp < 0) */ {
-        frac_part = (mantissa & FRAC_MASK2) >> -(exp + 1);
     }
 
+    if (decimals) {
+        double int_f = floor(af);
+        double frac_f = round((af - int_f) * pow10v[decimals]);
+
+        int_part = (Uint64)int_f;
+        frac_part = (Uint64)frac_f;
+
+        if (frac_f >= pow10v[decimals]) {
+            /* rounding overflow carry into int_part */
+            int_part++;
+            frac_part = 0;
+        }
+
+        do {
+            Uint64 n;
+            if (!frac_part) {
+                do {
+                    *p++ = '0';
+                } while (--decimals);
+                break;
+            }
+            n = frac_part / 10;
+            *p++ = (char)((frac_part - n*10) + '0');
+            frac_part = n;
+        } while (--decimals);
+
+        *p++ = '.';
+    }
+    else
+        int_part = (Uint64)round_int64(af);
+
     if (!int_part) {
-        if (neg)
-            *p++ = '-';
         *p++ = '0';
     } else {
-        int ret, i, n;
-        while (int_part != 0) {
-            long long j = int_part / 10;
-            *p++ = (char)(int_part - ((j << 3) + (j << 1)) + '0');
-            int_part = j;
-        }
-        if (neg)
-            *p++ = '-';
-        /* Reverse string */
-        ret = p - buffer;
-        for (i = 0, n = ret/2; i < n; i++) {
-            int  j = ret - i - 1;
-            char c = buffer[i];
+        do {
+            Uint64 n = int_part / 10;
+            *p++ = (char)((int_part - n*10) + '0');
+            int_part = n;
+        } while (int_part);
+    }
+    if (neg)
+        *p++ = '-';
+
+    {/* Reverse string */
+        int i = 0;
+        int j = p - buffer - 1;
+        for ( ; i < j; i++, j--) {
+            char tmp = buffer[i];
             buffer[i] = buffer[j];
-            buffer[j] = c;
+            buffer[j] = tmp;
         }
     }
 
-    if (decimals > 0) {
-        int i;
-        *p++ = '.';
-
-        max = buffer_size - (p - buffer) - 1 /* leave room for trailing '\0' */;
-
-        if (decimals > max)
-            return -1;  /* the number is not large enough to fit in the buffer */
-
-        max = decimals;
-
-        for (i = 0; i < max; i++) {
-            /* frac_part *= 10; */
-            frac_part = (frac_part << 3) + (frac_part << 1);
-
-            *p++ = (char)((frac_part >> (FRAC_SIZE + 1)) + '0');
-            frac_part &= FRAC_MASK2;
-        }
-
-        /* Delete trailing zeroes */
-        if (compact)
-            p = find_first_trailing_zero(p);
-    }
-
+    /* Delete trailing zeroes */
+    if (compact)
+        p = find_first_trailing_zero(p);
     *p = '\0';
     return p - buffer;
 }
