@@ -50,7 +50,6 @@
 #endif
 
 #define ERTS_WANT_BREAK_HANDLING
-#define ERTS_WANT_GOT_SIGUSR1
 #define WANT_NONBLOCKING    /* must define this to pull in defs from sys.h */
 #include "sys.h"
 #include "erl_thr_progress.h"
@@ -59,15 +58,10 @@
 #define __DARWIN__ 1
 #endif
 
-#ifdef USE_THREADS
 #include "erl_threads.h"
-#endif
 
 #include "erl_mseg.h"
 
-extern char **environ;
-erts_smp_rwmtx_t environ_rwmtx;
-
 #define MAX_VSIZE 16		/* Max number of entries allowed in an I/O
 				 * vector sock_sendv().
 				 */
@@ -80,7 +74,7 @@ erts_smp_rwmtx_t environ_rwmtx;
 
 #include "erl_check_io.h"
 #include "erl_cpu_topology.h"
-
+#include "erl_osenv.h"
 extern int  driver_interrupt(int, int);
 extern void do_break(void);
 
@@ -95,35 +89,19 @@ extern void erts_sys_init_float(void);
 static int debug_log = 0;
 #endif
 
-#ifdef ERTS_SMP
-erts_smp_atomic32_t erts_got_sigusr1;
-#define ERTS_SET_GOT_SIGUSR1 \
-  erts_smp_atomic32_set_mb(&erts_got_sigusr1, 1)
-#define ERTS_UNSET_GOT_SIGUSR1 \
-  erts_smp_atomic32_set_mb(&erts_got_sigusr1, 0)
-static erts_smp_atomic32_t have_prepared_crash_dump;
+static erts_atomic32_t have_prepared_crash_dump;
 #define ERTS_PREPARED_CRASH_DUMP \
-  ((int) erts_smp_atomic32_xchg_nob(&have_prepared_crash_dump, 1))
-#else
-volatile int erts_got_sigusr1;
-#define ERTS_SET_GOT_SIGUSR1 (erts_got_sigusr1 = 1)
-#define ERTS_UNSET_GOT_SIGUSR1 (erts_got_sigusr1 = 0)
-static volatile int have_prepared_crash_dump;
-#define ERTS_PREPARED_CRASH_DUMP \
-  (have_prepared_crash_dump++)
-#endif
+  ((int) erts_atomic32_xchg_nob(&have_prepared_crash_dump, 1))
 
-erts_smp_atomic_t sys_misc_mem_sz;
+erts_atomic_t sys_misc_mem_sz;
 
-#if defined(ERTS_SMP)
-static void smp_sig_notify(char c);
+static void smp_sig_notify(int signum);
 static int sig_notify_fds[2] = {-1, -1};
 
 #ifdef ERTS_SYS_SUSPEND_SIGNAL
 static int sig_suspend_fds[2] = {-1, -1};
 #endif
 
-#endif
 
 jmp_buf erts_sys_sigsegv_jmp;
 
@@ -137,149 +115,36 @@ static int max_files = -1;
 /* 
  * a few variables used by the break handler 
  */
-#ifdef ERTS_SMP
-erts_smp_atomic32_t erts_break_requested;
+erts_atomic32_t erts_break_requested;
 #define ERTS_SET_BREAK_REQUESTED \
-  erts_smp_atomic32_set_nob(&erts_break_requested, (erts_aint32_t) 1)
+  erts_atomic32_set_nob(&erts_break_requested, (erts_aint32_t) 1)
 #define ERTS_UNSET_BREAK_REQUESTED \
-  erts_smp_atomic32_set_nob(&erts_break_requested, (erts_aint32_t) 0)
-#else
-volatile int erts_break_requested = 0;
-#define ERTS_SET_BREAK_REQUESTED (erts_break_requested = 1)
-#define ERTS_UNSET_BREAK_REQUESTED (erts_break_requested = 0)
-#endif
+  erts_atomic32_set_nob(&erts_break_requested, (erts_aint32_t) 0)
+
+
 /* set early so the break handler has access to initial mode */
 static struct termios initial_tty_mode;
 static int replace_intr = 0;
 /* assume yes initially, ttsl_init will clear it */
-int using_oldshell = 1; 
+int using_oldshell = 1;
 
-#ifdef ERTS_ENABLE_KERNEL_POLL
-
-int erts_use_kernel_poll = 0;
-
-struct {
-    int (*select)(ErlDrvPort, ErlDrvEvent, int, int);
-    int (*event)(ErlDrvPort, ErlDrvEvent, ErlDrvEventData);
-    void (*check_io_as_interrupt)(void);
-    void (*check_io_interrupt)(int);
-    void (*check_io_interrupt_tmd)(int, ErtsMonotonicTime);
-    void (*check_io)(int);
-    Uint (*size)(void);
-    Eterm (*info)(void *);
-    int (*check_io_debug)(ErtsCheckIoDebugInfo *);
-} io_func = {0};
-
-
-int
-driver_select(ErlDrvPort port, ErlDrvEvent event, int mode, int on)
+UWord
+erts_sys_get_page_size(void)
 {
-    return (*io_func.select)(port, event, mode, on);
-}
-
-int
-driver_event(ErlDrvPort port, ErlDrvEvent event, ErlDrvEventData event_data)
-{
-    return (*io_func.event)(port, event, event_data);
-}
-
-Eterm erts_check_io_info(void *p)
-{
-    return (*io_func.info)(p);
-}
-
-int
-erts_check_io_debug(ErtsCheckIoDebugInfo *ip)
-{
-    return (*io_func.check_io_debug)(ip);
-}
-
-
-static void
-init_check_io(void)
-{
-    if (erts_use_kernel_poll) {
-	io_func.select			= driver_select_kp;
-	io_func.event			= driver_event_kp;
-#ifdef ERTS_POLL_NEED_ASYNC_INTERRUPT_SUPPORT
-	io_func.check_io_as_interrupt	= erts_check_io_async_sig_interrupt_kp;
-#endif
-	io_func.check_io_interrupt	= erts_check_io_interrupt_kp;
-	io_func.check_io_interrupt_tmd	= erts_check_io_interrupt_timed_kp;
-	io_func.check_io		= erts_check_io_kp;
-	io_func.size			= erts_check_io_size_kp;
-	io_func.info			= erts_check_io_info_kp;
-	io_func.check_io_debug		= erts_check_io_debug_kp;
-	erts_init_check_io_kp();
-	max_files = erts_check_io_max_files_kp();
-    }
-    else {
-	io_func.select			= driver_select_nkp;
-	io_func.event			= driver_event_nkp;
-#ifdef ERTS_POLL_NEED_ASYNC_INTERRUPT_SUPPORT
-	io_func.check_io_as_interrupt	= erts_check_io_async_sig_interrupt_nkp;
-#endif
-	io_func.check_io_interrupt	= erts_check_io_interrupt_nkp;
-	io_func.check_io_interrupt_tmd	= erts_check_io_interrupt_timed_nkp;
-	io_func.check_io		= erts_check_io_nkp;
-	io_func.size			= erts_check_io_size_nkp;
-	io_func.info			= erts_check_io_info_nkp;
-	io_func.check_io_debug		= erts_check_io_debug_nkp;
-	erts_init_check_io_nkp();
-	max_files = erts_check_io_max_files_nkp();
-    }
-}
-
-#ifdef ERTS_POLL_NEED_ASYNC_INTERRUPT_SUPPORT
-#define ERTS_CHK_IO_AS_INTR()	(*io_func.check_io_as_interrupt)()
+#if defined(_SC_PAGESIZE)
+    return (UWord) sysconf(_SC_PAGESIZE);
+#elif defined(HAVE_GETPAGESIZE)
+    return (UWord) getpagesize();
 #else
-#define ERTS_CHK_IO_AS_INTR()	(*io_func.check_io_interrupt)(1)
+    return (UWord) 4*1024; /* Guess 4 KB */
 #endif
-#define ERTS_CHK_IO_INTR	(*io_func.check_io_interrupt)
-#define ERTS_CHK_IO_INTR_TMD	(*io_func.check_io_interrupt_tmd)
-#define ERTS_CHK_IO		(*io_func.check_io)
-#define ERTS_CHK_IO_SZ		(*io_func.size)
-
-#else /* !ERTS_ENABLE_KERNEL_POLL */
-
-static void
-init_check_io(void)
-{
-    erts_init_check_io();
-    max_files = erts_check_io_max_files();
 }
 
-#ifdef ERTS_POLL_NEED_ASYNC_INTERRUPT_SUPPORT
-#define ERTS_CHK_IO_AS_INTR()	erts_check_io_async_sig_interrupt()
-#else
-#define ERTS_CHK_IO_AS_INTR()	erts_check_io_interrupt(1)
-#endif
-#define ERTS_CHK_IO_INTR	erts_check_io_interrupt
-#define ERTS_CHK_IO_INTR_TMD	erts_check_io_interrupt_timed
-#define ERTS_CHK_IO		erts_check_io
-#define ERTS_CHK_IO_SZ		erts_check_io_size
-
-#endif
-
-void
-erts_sys_schedule_interrupt(int set)
-{
-    ERTS_CHK_IO_INTR(set);
-}
-
-#ifdef ERTS_SMP
-void
-erts_sys_schedule_interrupt_timed(int set, ErtsMonotonicTime timeout_time)
-{
-    ERTS_CHK_IO_INTR_TMD(set, timeout_time);
-}
-#endif
-
 Uint
 erts_sys_misc_mem_sz(void)
 {
-    Uint res = ERTS_CHK_IO_SZ();
-    res += erts_smp_atomic_read_mb(&sys_misc_mem_sz);
+    Uint res = erts_check_io_size();
+    res += erts_atomic_read_mb(&sys_misc_mem_sz);
     return res;
 }
 
@@ -304,7 +169,6 @@ MALLOC_USE_HASH(1);
 #endif
 #endif
 
-#ifdef USE_THREADS
 
 #ifdef ERTS_THR_HAVE_SIG_FUNCS
 
@@ -383,19 +247,15 @@ thr_create_prepare_child(void *vtcdp)
     erts_sched_bind_atthrcreate_child(tcdp->sched_bind_data);
 }
 
-#endif /* #ifdef USE_THREADS */
 
 void
 erts_sys_pre_init(void)
 {
-#ifdef USE_THREADS
     erts_thr_init_data_t eid = ERTS_THR_INIT_DATA_DEF_INITER;
-#endif
 
     erts_printf_add_cr_to_stdout = 1;
     erts_printf_add_cr_to_stderr = 1;
 
-#ifdef USE_THREADS
 
     eid.thread_create_child_func = thr_create_prepare_child;
     /* Before creation in parent */
@@ -403,42 +263,29 @@ erts_sys_pre_init(void)
     /* After creation in parent */
     eid.thread_create_parent_func = thr_create_cleanup,
 
-#ifdef ERTS_THR_HAVE_SIG_FUNCS
-    sigemptyset(&thr_create_sigmask);
-    sigaddset(&thr_create_sigmask, SIGINT);   /* block interrupt */
-    sigaddset(&thr_create_sigmask, SIGTERM);  /* block terminate signal */
-    sigaddset(&thr_create_sigmask, SIGUSR1);  /* block user defined signal */
+#ifdef ERTS_ENABLE_LOCK_COUNT
+    erts_lcnt_pre_thr_init();
 #endif
 
     erts_thr_init(&eid);
 
+#ifdef ERTS_ENABLE_LOCK_COUNT
+    erts_lcnt_post_thr_init();
+#endif
+
 #ifdef ERTS_ENABLE_LOCK_CHECK
     erts_lc_init();
 #endif
 
-#ifdef ERTS_ENABLE_LOCK_COUNT
-    erts_lcnt_init();
-#endif
-
-#endif /* USE_THREADS */
 
     erts_init_sys_time_sup();
 
-#ifdef USE_THREADS
 
-#ifdef ERTS_SMP
-    erts_smp_atomic32_init_nob(&erts_break_requested, 0);
-    erts_smp_atomic32_init_nob(&erts_got_sigusr1, 0);
-    erts_smp_atomic32_init_nob(&have_prepared_crash_dump, 0);
-#else
-    erts_break_requested = 0;
-    erts_got_sigusr1 = 0;
-    have_prepared_crash_dump = 0;
-#endif
+    erts_atomic32_init_nob(&erts_break_requested, 0);
+    erts_atomic32_init_nob(&have_prepared_crash_dump, 0);
 
-#endif /* USE_THREADS */
 
-    erts_smp_atomic_init_nob(&sys_misc_mem_sz, 0);
+    erts_atomic_init_nob(&sys_misc_mem_sz, 0);
 
     {
       /*
@@ -501,10 +348,8 @@ SIGFUNC sys_signal(int sig, SIGFUNC func)
     return(oact.sa_handler);
 }
 
-#ifdef USE_THREADS
 #undef  sigprocmask
 #define sigprocmask erts_thr_sigmask
-#endif
 
 void sys_sigblock(int sig)
 {
@@ -524,11 +369,13 @@ void sys_sigrelease(int sig)
     sigprocmask(SIG_UNBLOCK, &mask, (sigset_t *)NULL);
 }
 
+#ifdef ERTS_HAVE_TRY_CATCH
 void erts_sys_sigsegv_handler(int signo) {
     if (signo == SIGSEGV) {
         longjmp(erts_sys_sigsegv_jmp, 1);
     }
 }
+#endif
 
 /*
  * Function returns 1 if we can read from all values in between
@@ -604,10 +451,10 @@ prepare_crash_dump(int secs)
     close(crashdump_companion_cube_fd);
 
     envsz = sizeof(env);
-    i = erts_sys_getenv__("ERL_CRASH_DUMP_NICE", env, &envsz);
+    i = erts_sys_explicit_8bit_getenv("ERL_CRASH_DUMP_NICE", env, &envsz);
     if (i >= 0) {
 	int nice_val;
-	nice_val = i != 0 ? 0 : atoi(env);
+	nice_val = i != 1 ? 0 : atoi(env);
 	if (nice_val > 39) {
 	    nice_val = 39;
 	}
@@ -624,6 +471,28 @@ int erts_sys_prepare_crash_dump(int secs)
     return prepare_crash_dump(secs);
 }
 
+static void signal_notify_requested(Eterm type) {
+    Process* p = NULL;
+    Eterm msg, *hp;
+    ErtsProcLocks locks = 0;
+    ErlOffHeap *ohp;
+
+    Eterm id = erts_whereis_name_to_id(NULL, am_erl_signal_server);
+
+    if ((p = (erts_pid2proc_opt(NULL, 0, id, 0, ERTS_P2P_FLG_INC_REFC))) != NULL) {
+        ErtsMessage *msgp = erts_alloc_message_heap(p, &locks, 3, &hp, &ohp);
+
+        /* erl_signal_server ! {notify, sighup} */
+        msg = TUPLE2(hp, am_notify, type);
+        erts_queue_message(p, locks, msgp, msg, am_system);
+
+        if (locks)
+            erts_proc_unlock(p, locks);
+        erts_proc_dec_refc(p);
+    }
+}
+
+
 static ERTS_INLINE void
 break_requested(void)
 {
@@ -631,84 +500,17 @@ break_requested(void)
    * just set a flag - checked for and handled by
    * scheduler threads erts_check_io() (not signal handler).
    */
-#ifdef DEBUG			
-  fprintf(stderr,"break!\n");
-#endif
   if (ERTS_BREAK_REQUESTED)
       erts_exit(ERTS_INTR_EXIT, "");
 
   ERTS_SET_BREAK_REQUESTED;
-  ERTS_CHK_IO_AS_INTR(); /* Make sure we don't sleep in poll */
+  /* Wake aux thread to get handle break */
+  erts_aux_thread_poke();
 }
 
-/* set up signal handlers for break and quit */
-#if (defined(SIG_SIGSET) || defined(SIG_SIGNAL))
-static RETSIGTYPE request_break(void)
-#else
 static RETSIGTYPE request_break(int signum)
-#endif
 {
-#ifdef ERTS_SMP
-    smp_sig_notify('I');
-#else
-    break_requested();
-#endif
-}
-
-static void stop_requested(void) {
-    Process* p = NULL;
-    Eterm msg, *hp;
-    ErtsProcLocks locks = 0;
-    ErlOffHeap *ohp;
-    Eterm id = erts_whereis_name_to_id(NULL, am_init);
-
-    if ((p = (erts_pid2proc_opt(NULL, 0, id, 0, ERTS_P2P_FLG_INC_REFC))) != NULL) {
-        ErtsMessage *msgp = erts_alloc_message_heap(p, &locks, 3, &hp, &ohp);
-
-        /* init ! {stop,stop} */
-        msg = TUPLE2(hp, am_stop, am_stop);
-        erts_queue_message(p, locks, msgp, msg, am_system);
-
-        if (locks)
-            erts_smp_proc_unlock(p, locks);
-        erts_proc_dec_refc(p);
-    }
-}
-
-#if (defined(SIG_SIGSET) || defined(SIG_SIGNAL))
-static RETSIGTYPE request_stop(void)
-#else
-static RETSIGTYPE request_stop(int signum)
-#endif
-{
-#ifdef ERTS_SMP
-    smp_sig_notify('S');
-#else
-    stop_requested();
-#endif
-}
-
-
-static ERTS_INLINE void
-sigusr1_exit(void)
-{
-    char env[21]; /* enough to hold any 64-bit integer */
-    size_t envsz;
-    int i, secs = -1;
-
-    /* We do this at interrupt level, since the main reason for
-     * wanting to generate a crash dump in this way is that the emulator
-     * is hung somewhere, so it won't be able to poll any flag we set here.
-     */
-    ERTS_SET_GOT_SIGUSR1;
-
-    envsz = sizeof(env);
-    if ((i = erts_sys_getenv_raw("ERL_CRASH_DUMP_SECONDS", env, &envsz)) >= 0) {
-	secs = i != 0 ? 0 : atoi(env);
-    }
-
-    prepare_crash_dump(secs);
-    erts_exit(ERTS_DUMP_EXIT, "Received SIGUSR1\n");
+    smp_sig_notify(signum);
 }
 
 #ifdef ETHR_UNUSABLE_SIGUSRX
@@ -731,19 +533,6 @@ sys_thr_resume(erts_tid_t tid) {
 }
 #endif
 
-#if (defined(SIG_SIGSET) || defined(SIG_SIGNAL))
-static RETSIGTYPE user_signal1(void)
-#else
-static RETSIGTYPE user_signal1(int signum)
-#endif
-{
-#ifdef ERTS_SMP
-   smp_sig_notify('1');
-#else
-   sigusr1_exit();
-#endif
-}
-
 #ifdef ERTS_SYS_SUSPEND_SIGNAL
 #if (defined(SIG_SIGSET) || defined(SIG_SIGNAL))
 static RETSIGTYPE suspend_signal(void)
@@ -763,31 +552,103 @@ static RETSIGTYPE suspend_signal(int signum)
 
 #endif /* #ifndef ETHR_UNUSABLE_SIGUSRX */
 
-static void
-quit_requested(void)
+/*
+ Signal      Action   Comment
+ ─────────────────────────────────────────────────────────────
+  SIGHUP     Term     Hangup detected on controlling terminal or death of controlling process
+ !SIGINT     Term     Interrupt from keyboard
+  SIGQUIT    Core     Quit from keyboard
+ !SIGILL     Core     Illegal Instruction
+  SIGABRT    Core     Abort signal from abort(3)
+ !SIGFPE     Core     Floating point exception
+ !SIGKILL    Term     Kill signal
+ !SIGSEGV    Core     Invalid memory reference
+ !SIGPIPE    Term     Broken pipe: write to pipe with no readers
+  SIGALRM    Term     Timer signal from alarm(2)
+  SIGTERM    Term     Termination signal
+  SIGUSR1    Term     User-defined signal 1
+  SIGUSR2    Term     User-defined signal 2
+ !SIGCHLD    Ign      Child stopped or terminated
+ !SIGCONT    Cont     Continue if stopped
+  SIGSTOP    Stop     Stop process
+  SIGTSTP    Stop     Stop typed at terminal
+ !SIGTTIN    Stop     Terminal input for background process
+ !SIGTTOU    Stop     Terminal output for background process
+*/
+
+
+static ERTS_INLINE int
+signalterm_to_signum(Eterm signal)
 {
-    erts_exit(ERTS_INTR_EXIT, "");
+    switch (signal) {
+    case am_sighup:  return SIGHUP;
+    /* case am_sigint:  return SIGINT; */
+    case am_sigquit: return SIGQUIT;
+    /* case am_sigill:  return SIGILL; */
+    case am_sigabrt: return SIGABRT;
+    /* case am_sigsegv: return SIGSEGV; */
+    case am_sigalrm: return SIGALRM;
+    case am_sigterm: return SIGTERM;
+    case am_sigusr1: return SIGUSR1;
+    case am_sigusr2: return SIGUSR2;
+    case am_sigchld: return SIGCHLD;
+    case am_sigstop: return SIGSTOP;
+    case am_sigtstp: return SIGTSTP;
+    default:         return 0;
+    }
 }
 
-#if (defined(SIG_SIGSET) || defined(SIG_SIGNAL))
-static RETSIGTYPE do_quit(void)
-#else
-static RETSIGTYPE do_quit(int signum)
-#endif
+static ERTS_INLINE Eterm
+signum_to_signalterm(int signum)
 {
-#ifdef ERTS_SMP
-    smp_sig_notify('Q');
-#else
-    quit_requested();
-#endif
+    switch (signum) {
+    case SIGHUP:  return am_sighup;
+    /* case SIGINT:  return am_sigint; */    /* ^c */
+    case SIGQUIT: return am_sigquit;   /* ^\ */
+    /* case SIGILL:  return am_sigill; */
+    case SIGABRT: return am_sigabrt;
+    /* case SIGSEGV: return am_sigsegv; */
+    case SIGALRM: return am_sigalrm;
+    case SIGTERM: return am_sigterm;
+    case SIGUSR1: return am_sigusr1;
+    case SIGUSR2: return am_sigusr2;
+    case SIGCHLD: return am_sigchld;
+    case SIGSTOP: return am_sigstop;
+    case SIGTSTP: return am_sigtstp;   /* ^z */
+    default:      return am_error;
+    }
+}
+
+static RETSIGTYPE generic_signal_handler(int signum)
+{
+    smp_sig_notify(signum);
+}
+
+int erts_set_signal(Eterm signal, Eterm type) {
+    int signum;
+    if ((signum = signalterm_to_signum(signal)) > 0) {
+        if (type == am_ignore) {
+            sys_signal(signum, SIG_IGN);
+        } else if (type == am_default) {
+            sys_signal(signum, SIG_DFL);
+        } else {
+            sys_signal(signum, generic_signal_handler);
+        }
+        return 1;
+    }
+    return 0;
 }
 
 /* Disable break */
 void erts_set_ignore_break(void) {
-    sys_signal(SIGINT,  SIG_IGN);
-    sys_signal(SIGTERM, SIG_IGN);
-    sys_signal(SIGQUIT, SIG_IGN);
-    sys_signal(SIGTSTP, SIG_IGN);
+    /*
+     * Ignore signals that can be sent to the VM by
+     * typing certain key combinations at the
+     * controlling terminal...
+     */
+    sys_signal(SIGINT,  SIG_IGN);       /* Ctrl-C */
+    sys_signal(SIGQUIT, SIG_IGN);       /* Ctrl-\ */
+    sys_signal(SIGTSTP, SIG_IGN);       /* Ctrl-Z */
 }
 
 /* Don't use ctrl-c for break handler but let it be 
@@ -797,11 +658,11 @@ void erts_replace_intr(void) {
 
   if (isatty(0)) {
     tcgetattr(0, &mode);
-    
+
     /* here's an example of how to replace ctrl-c with ctrl-u */
     /* mode.c_cc[VKILL] = 0;
        mode.c_cc[VINTR] = CKILL; */
-    
+
     mode.c_cc[VINTR] = 0;	/* disable ctrl-c */
     tcsetattr(0, TCSANOW, &mode);
     replace_intr = 1;
@@ -810,12 +671,11 @@ void erts_replace_intr(void) {
 
 void init_break_handler(void)
 {
-   sys_signal(SIGINT, request_break);
-   sys_signal(SIGTERM, request_stop);
+   sys_signal(SIGINT,  request_break);
 #ifndef ETHR_UNUSABLE_SIGUSRX
-   sys_signal(SIGUSR1, user_signal1);
+   sys_signal(SIGUSR1, generic_signal_handler);
 #endif /* #ifndef ETHR_UNUSABLE_SIGUSRX */
-   sys_signal(SIGQUIT, do_quit);
+   sys_signal(SIGQUIT, generic_signal_handler);
 }
 
 void sys_init_suspend_handler(void)
@@ -825,9 +685,15 @@ void sys_init_suspend_handler(void)
 #endif
 }
 
+void
+erts_sys_unix_later_init(void)
+{
+    sys_signal(SIGTERM, generic_signal_handler);
+}
+
 int sys_max_files(void)
 {
-   return(max_files);
+   return max_files;
 }
 
 /************************** OS info *******************************/
@@ -855,10 +721,7 @@ get_number(char **str_ptr)
     }
 }
 
-void
-os_flavor(char* namebuf, 	/* Where to return the name. */
-	  unsigned size) 	/* Size of name buffer. */
-{
+void os_flavor(char* namebuf, unsigned size) {
     struct utsname uts;		/* Information about the system. */
     char* s;
 
@@ -871,63 +734,29 @@ os_flavor(char* namebuf, 	/* Where to return the name. */
     strcpy(namebuf, uts.sysname);
 }
 
-void
-os_version(pMajor, pMinor, pBuild)
-int* pMajor;			/* Pointer to major version. */
-int* pMinor;			/* Pointer to minor version. */
-int* pBuild;			/* Pointer to build number. */
-{
+void os_version(int *pMajor, int *pMinor, int *pBuild) {
     struct utsname uts;		/* Information about the system. */
     char* release;		/* Pointer to the release string:
-				 * X.Y or X.Y.Z.
-				 */
+				 * X.Y or X.Y.Z.  */
 
     (void) uname(&uts);
     release = uts.release;
-    *pMajor = get_number(&release);
-    *pMinor = get_number(&release);
-    *pBuild = get_number(&release);
-}
-
-void init_getenv_state(GETENV_STATE *state)
-{
-   erts_smp_rwmtx_rlock(&environ_rwmtx);
-   *state = NULL;
-}
-
-char *getenv_string(GETENV_STATE *state0)
-{
-   char **state = (char **) *state0;
-   char *cp;
-
-   ERTS_SMP_LC_ASSERT(erts_smp_lc_rwmtx_is_rlocked(&environ_rwmtx));
-
-   if (state == NULL)
-      state = environ;
-
-   cp = *state++;
-   *state0 = (GETENV_STATE) state;
-
-   return cp;
-}
-
-void fini_getenv_state(GETENV_STATE *state)
-{
-   *state = NULL;
-   erts_smp_rwmtx_runlock(&environ_rwmtx);
+    *pMajor = get_number(&release); /* Pointer to major version. */
+    *pMinor = get_number(&release); /* Pointer to minor version. */
+    *pBuild = get_number(&release); /* Pointer to build number. */
 }
 
 void erts_do_break_handling(void)
 {
     struct termios temp_mode;
     int saved = 0;
-    
+
     /*
      * Most functions that do_break() calls are intentionally not thread safe;
      * therefore, make sure that all threads but this one are blocked before
      * proceeding!
      */
-    erts_smp_thr_progress_block();
+    erts_thr_progress_block();
 
     /* during break we revert to initial settings */
     /* this is done differently for oldshell */
@@ -939,14 +768,14 @@ void erts_do_break_handling(void)
       tcsetattr(0,TCSANOW,&initial_tty_mode);
       saved = 1;
     }
-    
+
     /* call the break handling function, reset the flag */
     do_break();
 
     ERTS_UNSET_BREAK_REQUESTED;
 
     fflush(stdout);
-    
+
     /* after break we go back to saved settings */
     if (using_oldshell && !replace_intr) {
       SET_NONBLOCKING(1);
@@ -955,9 +784,10 @@ void erts_do_break_handling(void)
       tcsetattr(0,TCSANOW,&temp_mode);
     }
 
-    erts_smp_thr_progress_unblock();
+    erts_thr_progress_unblock();
 }
 
+
 /* Fills in the systems representation of the jam/beam process identifier.
 ** The Pid is put in STRING representation in the supplied buffer,
 ** no interpretatione of this should be done by the rest of the
@@ -969,122 +799,13 @@ void sys_get_pid(char *buffer, size_t buffer_size){
     erts_snprintf(buffer, buffer_size, "%lu",(unsigned long) p);
 }
 
-int
-erts_sys_putenv_raw(char *key, char *value) {
-    return erts_sys_putenv(key, value);
-}
-int
-erts_sys_putenv(char *key, char *value)
-{
-    int res;
-    char *env;
-    Uint need = strlen(key) + strlen(value) + 2;
 
-#ifdef HAVE_COPYING_PUTENV
-    env = erts_alloc(ERTS_ALC_T_TMP, need);
-#else
-    env = erts_alloc(ERTS_ALC_T_PUTENV_STR, need);
-    erts_smp_atomic_add_nob(&sys_misc_mem_sz, need);
-#endif
-    strcpy(env,key);
-    strcat(env,"=");
-    strcat(env,value);
-    erts_smp_rwmtx_rwlock(&environ_rwmtx);
-    res = putenv(env);
-    erts_smp_rwmtx_rwunlock(&environ_rwmtx);
-#ifdef HAVE_COPYING_PUTENV
-    erts_free(ERTS_ALC_T_TMP, env);
-#endif
-    return res;
-}
-
-int
-erts_sys_getenv__(char *key, char *value, size_t *size)
-{
-    int res;
-    char *orig_value = getenv(key);
-    if (!orig_value)
-	res = -1;
-    else {
-	size_t len = sys_strlen(orig_value);
-	if (len >= *size) {
-	    *size = len + 1;
-	    res = 1;
-	}
-	else {
-	    *size = len;
-	    sys_memcpy((void *) value, (void *) orig_value, len+1);
-	    res = 0;
-	}
-    }
-    return res;
-}
-
-int
-erts_sys_getenv_raw(char *key, char *value, size_t *size) {
-    return erts_sys_getenv(key, value, size);
-}
-
-/*
- * erts_sys_getenv
- * returns:
- *  -1, if environment key is not set with a value
- *   0, if environment key is set and value fits into buffer size
- *   1, if environment key is set but does not fit into buffer size
- *      size is set with the needed buffer size value
- */
-
-int
-erts_sys_getenv(char *key, char *value, size_t *size)
-{
-    int res;
-    erts_smp_rwmtx_rlock(&environ_rwmtx);
-    res = erts_sys_getenv__(key, value, size);
-    erts_smp_rwmtx_runlock(&environ_rwmtx);
-    return res;
-}
-
-int
-erts_sys_unsetenv(char *key)
-{
-    int res;
-    erts_smp_rwmtx_rwlock(&environ_rwmtx);
-    res = unsetenv(key);
-    erts_smp_rwmtx_rwunlock(&environ_rwmtx);
-    return res;
-}
-
-void
-sys_init_io(void)
-{
-}
-
-#if (0) /* unused? */
-static int write_fill(fd, buf, len)
-int fd, len;
-char *buf;
-{
-    int i, done = 0;
-    
-    do {
-	if ((i = write(fd, buf+done, len-done)) < 0) {
-	    if (errno != EINTR)
-		return (i);
-	    i = 0;
-	}
-	done += i;
-    } while (done < len);
-    return (len);
-}
-#endif
+void sys_init_io(void) { }
+void erts_sys_alloc_init(void) { }
 
 extern const char pre_loaded_code[];
 extern Preload pre_loaded[];
 
-void erts_sys_alloc_init(void)
-{
-}
-
 #if ERTS_HAVE_ERTS_SYS_ALIGNED_ALLOC
 void *erts_sys_aligned_alloc(UWord alignment, UWord size)
 {
@@ -1185,9 +906,7 @@ void sys_preload_end(Preload* p)
    Here we assume that all schedulers are stopped so that erl_poll
    does not interfere with the select below.
 */
-int sys_get_key(fd)
-int fd;
-{
+int sys_get_key(int fd) {
     int c, ret;
     unsigned char rbuf[64];
     fd_set fds;
@@ -1206,29 +925,18 @@ int fd;
         if (c <= 0)
             return c;
     }
-
-    return rbuf[0]; 
+    return rbuf[0];
 }
 
 
 extern int erts_initialized;
 void
 erl_assert_error(const char* expr, const char* func, const char* file, int line)
-{   
+{
     fflush(stdout);
     fprintf(stderr, "%s:%d:%s() Assertion failed: %s\n",
             file, line, func, expr);
     fflush(stderr);
-#if !defined(ERTS_SMP) && 0
-    /* Writing a crashdump from a failed assertion when smp support
-     * is enabled almost a guaranteed deadlocking, don't even bother.
-     *
-     * It could maybe be useful (but I'm not convinced) to write the
-     * crashdump if smp support is disabled...
-     */
-    if (erts_initialized)
-	erl_crash_dump(file, line, "Assertion failed: %s\n", expr);
-#endif
     abort();
 }
 
@@ -1239,7 +947,7 @@ erl_debug(char* fmt, ...)
 {
     char sbuf[1024];		/* Temporary buffer. */
     va_list va;
-    
+
     if (debug_log) {
 	va_start(va, fmt);
 	vsprintf(sbuf, fmt, va);
@@ -1250,32 +958,17 @@ erl_debug(char* fmt, ...)
 
 #endif /* DEBUG */
 
-/*
- * Called from schedule() when it runs out of runnable processes,
- * or when Erlang code has performed INPUT_REDUCTIONS reduction
- * steps. runnable == 0 iff there are no runnable Erlang processes.
- */
-void
-erl_sys_schedule(int runnable)
-{
-    ERTS_CHK_IO(!runnable);
-    ERTS_SMP_LC_ASSERT(!erts_thr_progress_is_blocking());
-}
-
-
-#ifdef ERTS_SMP
-
-static erts_smp_tid_t sig_dispatcher_tid;
+static erts_tid_t sig_dispatcher_tid;
 
 static void
-smp_sig_notify(char c)
+smp_sig_notify(int signum)
 {
     int res;
     do {
 	/* write() is async-signal safe (according to posix) */
-	res = write(sig_notify_fds[1], &c, 1);
+	res = write(sig_notify_fds[1], &signum, sizeof(int));
     } while (res < 0 && errno == EINTR);
-    if (res != 1) {
+    if (res != sizeof(int)) {
 	char msg[] =
 	    "smp_sig_notify(): Failed to notify signal-dispatcher thread "
 	    "about received signal";
@@ -1291,60 +984,55 @@ signal_dispatcher_thread_func(void *unused)
     erts_lc_set_thread_name("signal_dispatcher");
 #endif
     while (1) {
-	char buf[32];
-	int res, i;
+        union {int signum; char buf[4];} sb;
+        Eterm signal;
+	int res, i = 0;
 	/* Block on read() waiting for a signal notification to arrive... */
-	res = read(sig_notify_fds[0], (void *) &buf[0], 32);
+
+        do {
+            res = read(sig_notify_fds[0], (void *) &sb.buf[i], sizeof(int) - i);
+            i += res > 0 ? res : 0;
+        } while ((i < sizeof(int) && res >= 0) || (res < 0 && errno == EINTR));
+
 	if (res < 0) {
-	    if (errno == EINTR)
-		continue;
 	    erts_exit(ERTS_ABORT_EXIT,
 		     "signal-dispatcher thread got unexpected error: %s (%d)\n",
 		     erl_errno_id(errno),
 		     errno);
 	}
-	for (i = 0; i < res; i++) {
-	    /*
-	     * NOTE 1: The signal dispatcher thread should not do work
-	     *         that takes a substantial amount of time (except
-	     *         perhaps in test and debug builds). It needs to
-	     *         be responsive, i.e, it should only dispatch work
-	     *         to other threads.
-	     *
-	     * NOTE 2: The signal dispatcher thread is not a blockable
-	     *         thread (i.e., not a thread managed by the
-	     *         erl_thr_progress module). This is intentional.
-	     *         We want to be able to interrupt writing of a crash
-	     *         dump by hitting C-c twice. Since it isn't a
-	     *         blockable thread it is important that it doesn't
-	     *         change the state of any data that a blocking thread
-	     *         expects to have exclusive access to (unless the
-	     *         signal dispatcher itself explicitly is blocking all
-	     *         blockable threads).
-	     */
-	    switch (buf[i]) {
-	    case 0: /* Emulator initialized */
+        /*
+         * NOTE 1: The signal dispatcher thread should not do work
+         *         that takes a substantial amount of time (except
+         *         perhaps in test and debug builds). It needs to
+         *         be responsive, i.e, it should only dispatch work
+         *         to other threads.
+         *
+         * NOTE 2: The signal dispatcher thread is not a blockable
+         *         thread (i.e., not a thread managed by the
+         *         erl_thr_progress module). This is intentional.
+         *         We want to be able to interrupt writing of a crash
+         *         dump by hitting C-c twice. Since it isn't a
+         *         blockable thread it is important that it doesn't
+         *         change the state of any data that a blocking thread
+         *         expects to have exclusive access to (unless the
+         *         signal dispatcher itself explicitly is blocking all
+         *         blockable threads).
+         */
+        switch (sb.signum) {
+            case 0: continue;
+            case SIGINT:
+                break_requested();
                 break;
-	    case 'S': /* SIGTERM */
-		stop_requested();
-		break;
-	    case 'I': /* SIGINT */
-		break_requested();
-		break;
-	    case 'Q': /* SIGQUIT */
-		quit_requested();
-		break;
-	    case '1': /* SIGUSR1 */
-		sigusr1_exit();
-		break;
-	    default:
-		erts_exit(ERTS_ABORT_EXIT,
-			 "signal-dispatcher thread received unknown "
-			 "signal notification: '%c'\n",
-			 buf[i]);
-	    }
-	}
-	ERTS_SMP_LC_ASSERT(!erts_thr_progress_is_blocking());
+            default:
+                if ((signal = signum_to_signalterm(sb.signum)) == am_error) {
+                    erts_exit(ERTS_ABORT_EXIT,
+                            "signal-dispatcher thread received unknown "
+                            "signal notification: '%d'\n",
+                            sb.signum);
+                }
+                signal_notify_requested(signal);
+        }
+        ERTS_LC_ASSERT(!erts_thr_progress_is_blocking());
     }
     return NULL;
 }
@@ -1352,7 +1040,7 @@ signal_dispatcher_thread_func(void *unused)
 static void
 init_smp_sig_notify(void)
 {
-    erts_smp_thr_opts_t thr_opts = ERTS_SMP_THR_OPTS_DEFAULT_INITER;
+    erts_thr_opts_t thr_opts = ERTS_THR_OPTS_DEFAULT_INITER;
     thr_opts.detached = 1;
     thr_opts.name = "sys_sig_dispatcher";
 
@@ -1364,7 +1052,7 @@ init_smp_sig_notify(void)
     }
 
     /* Start signal handler thread */
-    erts_smp_thr_create(&sig_dispatcher_tid,
+    erts_thr_create(&sig_dispatcher_tid,
 			signal_dispatcher_thread_func,
 			NULL,
 			&thr_opts);
@@ -1387,9 +1075,9 @@ init_smp_sig_suspend(void) {
 int erts_darwin_main_thread_pipe[2];
 int erts_darwin_main_thread_result_pipe[2];
 
-static void initialize_darwin_main_thread_pipes(void) 
+static void initialize_darwin_main_thread_pipes(void)
 {
-    if (pipe(erts_darwin_main_thread_pipe) < 0 || 
+    if (pipe(erts_darwin_main_thread_pipe) < 0 ||
 	pipe(erts_darwin_main_thread_result_pipe) < 0) {
 	erts_exit(ERTS_ERROR_EXIT,"Fatal error initializing Darwin main thread stealing");
     }
@@ -1402,12 +1090,12 @@ erts_sys_main_thread(void)
     erts_thread_disable_fpe();
 #ifdef __DARWIN__
     initialize_darwin_main_thread_pipes();
-#endif
+#else
     /* Become signal receiver thread... */
 #ifdef ERTS_ENABLE_LOCK_CHECK
     erts_lc_set_thread_name("signal_receiver");
 #endif
-
+#endif
     smp_sig_notify(0); /* Notify initialized */
 
     /* Wait for a signal to arrive... */
@@ -1457,103 +1145,15 @@ erts_sys_main_thread(void)
     }
 }
 
-#endif /* ERTS_SMP */
-
-#ifdef ERTS_ENABLE_KERNEL_POLL /* get_value() is currently only used when
-				  kernel-poll is enabled */
-
-/* Get arg marks argument as handled by
-   putting NULL in argv */
-static char *
-get_value(char* rest, char** argv, int* ip)
-{
-    char *param = argv[*ip]+1;
-    argv[*ip] = NULL;
-    if (*rest == '\0') {
-	char *next = argv[*ip + 1];
-	if (next[0] == '-'
-	    && next[1] == '-'
-	    &&  next[2] == '\0') {
-	    erts_fprintf(stderr, "bad \"%s\" value: \n", param);
-	    erts_usage();
-	}
-	(*ip)++;
-	argv[*ip] = NULL;
-	return next;
-    }
-    return rest;
-}
-
-#endif /* ERTS_ENABLE_KERNEL_POLL */
-
 void
 erl_sys_args(int* argc, char** argv)
 {
-    int i, j;
-
-    erts_smp_rwmtx_init(&environ_rwmtx, "environ");
-
-    i = 1;
-
     ASSERT(argc && argv);
 
-    while (i < *argc) {
-	if(argv[i][0] == '-') {
-	    switch (argv[i][1]) {
-#ifdef ERTS_ENABLE_KERNEL_POLL
-	    case 'K': {
-		char *arg = get_value(argv[i] + 2, argv, &i);
-		if (strcmp("true", arg) == 0) {
-		    erts_use_kernel_poll = 1;
-		}
-		else if (strcmp("false", arg) == 0) {
-		    erts_use_kernel_poll = 0;
-		}
-		else {
-		    erts_fprintf(stderr, "bad \"K\" value: %s\n", arg);
-		    erts_usage();
-		}
-		break;
-	    }
-#endif
-	    case '-':
-		goto done_parsing;
-	    default:
-		break;
-	    }
-	}
-	i++;
-    }
+    max_files = erts_check_io_max_files();
 
- done_parsing:
-
-#ifdef ERTS_ENABLE_KERNEL_POLL
-    if (erts_use_kernel_poll) {
-	char no_kp[10];
-	size_t no_kp_sz = sizeof(no_kp);
-	int res = erts_sys_getenv_raw("ERL_NO_KERNEL_POLL", no_kp, &no_kp_sz);
-	if (res > 0
-	    || (res == 0
-		&& sys_strcmp("false", no_kp) != 0
-		&& sys_strcmp("FALSE", no_kp) != 0)) {
-	    erts_use_kernel_poll = 0;
-	}
-    }
-#endif
-
-    init_check_io();
-
-#ifdef ERTS_SMP
     init_smp_sig_notify();
     init_smp_sig_suspend();
-#endif
-
-    /* Handled arguments have been marked with NULL. Slide arguments
-       not handled towards the beginning of argv. */
-    for (i = 0, j = 0; i < *argc; i++) {
-	if (argv[i])
-	    argv[j++] = argv[i];
-    }
-    *argc = j;
 
+    erts_sys_env_init();
 }
