@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %% 
-%% Copyright Ericsson AB 2004-2017. All Rights Reserved.
+%% Copyright Ericsson AB 2004-2018. All Rights Reserved.
 %% 
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -53,6 +53,7 @@ suite() ->
 all() ->
     [
      {group, http},
+     {group, http_ipv6},
      {group, sim_http},
      {group, http_internal},
      {group, http_unix_socket},
@@ -64,7 +65,11 @@ all() ->
 groups() ->
     [
      {http, [], real_requests()},
-     {sim_http, [], only_simulated() ++ [process_leak_on_keepalive]},
+     {http_ipv6, [], [request_options]},
+     %% process_leak_on_keepalive is depending on stream_fun_server_close
+     %% and it shall be the last test case in the suite otherwise cookie
+     %% will fail.
+     {sim_http, [], only_simulated() ++ server_closing_connection() ++ [process_leak_on_keepalive]},
      {http_internal, [], real_requests_esi()},
      {http_unix_socket, [], simulated_unix_socket()},
      {https, [], real_requests()},
@@ -76,6 +81,7 @@ real_requests()->
     [
      head,
      get,
+     get_query_string,
      post,
      delete,
      post_stream,
@@ -141,13 +147,22 @@ only_simulated() ->
      redirect_found,
      redirect_see_other,
      redirect_temporary_redirect,
+     redirect_relative_uri,
      port_in_host_header,
      redirect_port_in_host_header,
      relaxed,
      multipart_chunks,
+     get_space,
+     delete_no_body,
      stream_fun_server_close
     ].
 
+server_closing_connection() ->
+    [
+     server_closing_connection_on_first_response,
+     server_closing_connection_on_second_response
+    ].
+
 misc() ->
     [
      server_does_not_exist,
@@ -201,6 +216,16 @@ init_per_group(http_unix_socket = Group, Config0) ->
             Port = server_start(Group, server_config(Group, Config)),
             [{port, Port} | Config]
     end;
+init_per_group(http_ipv6 = Group, Config0) ->
+    case is_ipv6_supported() of
+        true ->
+            start_apps(Group),
+            Config = proplists:delete(port, Config0),
+            Port = server_start(Group, server_config(Group, Config)),
+            [{port, Port} | Config];
+        false ->
+            {skip, "Host does not support IPv6"}
+     end;
 init_per_group(Group, Config0) ->
     start_apps(Group),
     Config = proplists:delete(port, Config0),
@@ -214,9 +239,36 @@ end_per_group(_, _Config) ->
     ok.
 
 do_init_per_group(Group, Config0) ->
-    Config = proplists:delete(port, Config0),
+    Config1 =
+        case Group of
+            https ->
+                init_ssl(Config0);
+            sim_https ->
+                init_ssl(Config0);
+            _ ->
+                Config0
+        end,
+    Config = proplists:delete(port, Config1),
     Port = server_start(Group, server_config(Group, Config)),
     [{port, Port} | Config].
+
+init_ssl(Config) ->
+    ClientFileBase = filename:join([proplists:get_value(priv_dir, Config), "client"]),
+    ServerFileBase = filename:join([proplists:get_value(priv_dir, Config), "server"]),
+    GenCertData =
+        public_key:pkix_test_data(#{server_chain => 
+                                        #{root => [{key, inets_test_lib:hardcode_rsa_key(1)}],
+                                          intermediates => [[{key, inets_test_lib:hardcode_rsa_key(2)}]],
+                                          peer => [{key, inets_test_lib:hardcode_rsa_key(3)}
+                                                  ]},
+                                    client_chain => 
+                                        #{root => [{key, inets_test_lib:hardcode_rsa_key(4)}],
+                                          intermediates => [[{key, inets_test_lib:hardcode_rsa_key(5)}]],
+                                          peer => [{key, inets_test_lib:hardcode_rsa_key(6)}]}}),
+
+    Conf = inets_test_lib:gen_pem_config_files(GenCertData, ClientFileBase, ServerFileBase),
+    [{ssl_conf, Conf} | Config].
+
 %%--------------------------------------------------------------------
 init_per_testcase(pipeline, Config) ->
     inets:start(httpc, [{profile, pipeline}]),
@@ -227,7 +279,7 @@ init_per_testcase(pipeline, Config) ->
 init_per_testcase(persistent_connection, Config) ->
     inets:start(httpc, [{profile, persistent}]),
     httpc:set_options([{keep_alive_timeout, 50000},
-		       {max_keep_alive_length, 3}], persistent_connection),
+		       {max_keep_alive_length, 3}], persistent),
 
     Config;
 init_per_testcase(wait_for_whole_response, Config) ->
@@ -246,10 +298,38 @@ end_per_testcase(pipeline, _Config) ->
     inets:stop(httpc, pipeline);
 end_per_testcase(persistent_connection, _Config) ->
     inets:stop(httpc, persistent);
+end_per_testcase(Case, Config)
+  when Case == server_closing_connection_on_first_response;
+       Case == server_closing_connection_on_second_response ->
+    %% Test case uses at most one session.  Ensure no leftover
+    %% sessions left behind.
+    {_, Status} = proplists:lookup(tc_status, Config),
+    ShallCleanup = case Status of
+                       ok -> true;
+                       {failed, _} -> true;
+                       {skipped, _} -> false
+                   end,
+    if ShallCleanup =:= true ->
+            httpc:request(url(group_name(Config), "/just_close.html", Config)),
+            ok;
+       true ->
+            ct:pal("Not cleaning up because test case status was ~p", [Status]),
+            ok
+    end;
 
 end_per_testcase(_Case, _Config) ->
     ok.
 
+is_ipv6_supported() ->
+    case gen_udp:open(0, [inet6]) of
+        {ok, Socket} ->
+            gen_udp:close(Socket),
+            true;
+        _ ->
+            false
+    end.
+
+
 %%--------------------------------------------------------------------
 %% Test Cases --------------------------------------------------------
 %%--------------------------------------------------------------------
@@ -271,6 +351,25 @@ get(Config) when is_list(Config) ->
 
     {ok, {{_,200,_}, [_ | _], BinBody}} =  httpc:request(get, Request, [], [{body_format, binary}]),
     true = is_binary(BinBody).
+
+
+get_query_string() ->
+    [{doc, "Test http get request with query string against local server"}].
+get_query_string(Config) when is_list(Config) ->
+    Request  = {url(group_name(Config), "/dummy.html?foo=bar", Config), []},
+    {ok, {{_,200,_}, [_ | _], Body = [_ | _]}} = httpc:request(get, Request, [], []),
+
+    inets_test_lib:check_body(Body).
+
+%%--------------------------------------------------------------------
+get_space() ->
+    [{"Test http get request with '%20' in the path of the URL."}].
+get_space(Config) when is_list(Config) ->
+    Request  = {url(group_name(Config), "/space%20.html", Config), []},
+    {ok, {{_,200,_}, [_ | _], Body = [_ | _]}} = httpc:request(get, Request, [], []),
+
+    inets_test_lib:check_body(Body).
+
 %%--------------------------------------------------------------------
 post() ->
     [{"Test http post request against local server. We do in this case "
@@ -604,7 +703,26 @@ redirect_temporary_redirect(Config) when is_list(Config) ->
     {ok, {{_,200,_}, [_ | _], [_|_]}}
 	= httpc:request(post, {URL307, [],"text/plain", "foobar"},
 			[], []).
+%%-------------------------------------------------------------------------
+redirect_relative_uri() ->
+    [{doc, "The server SHOULD generate a Location header field in the response "
+      "containing a preferred URI reference for the new permanent URI.  The user "
+      "agent MAY use the Location field value for automatic redirection.  The server's "
+      "response payload usually contains a short hypertext note with a "
+      "hyperlink to the new URI(s)."}].
+redirect_relative_uri(Config) when is_list(Config) ->
 
+    URL301 = url(group_name(Config), "/301_rel_uri.html", Config),
+
+    {ok, {{_,200,_}, [_ | _], [_|_]}}
+	= httpc:request(get, {URL301, []}, [], []),
+
+    {ok, {{_,200,_}, [_ | _], []}}
+	= httpc:request(head, {URL301, []}, [], []),
+
+    {ok, {{_,200,_}, [_ | _], [_|_]}}
+	= httpc:request(post, {URL301, [],"text/plain", "foobar"},
+			[], []).
 %%-------------------------------------------------------------------------
 redirect_loop() ->
     [{"doc, Test redirect loop detection"}].
@@ -1082,8 +1200,6 @@ remote_socket_close_async(Config) when is_list(Config) ->
 %%-------------------------------------------------------------------------
 
 process_leak_on_keepalive(Config) ->
-    {ok, ClosedSocket} = gen_tcp:listen(6666, [{active, false}]),
-    ok = gen_tcp:close(ClosedSocket),
     Request  = {url(group_name(Config), "/dummy.html", Config), []},
     HttpcHandlers0 = supervisor:which_children(httpc_handler_sup),
     {ok, {{_, 200, _}, _, Body}} = httpc:request(get, Request, [], []),
@@ -1095,11 +1211,10 @@ process_leak_on_keepalive(Config) ->
         ordsets:to_list(
           ordsets:subtract(ordsets:from_list(HttpcHandlers1),
                            ordsets:from_list(HttpcHandlers0))),
-    sys:replace_state(
-      Pid, fun (State) ->
-                   Session = element(3, State),
-                   setelement(3, State, Session#session{socket=ClosedSocket})
-           end),
+    State = sys:get_state(Pid),
+    #session{socket=Socket} = element(3, State),
+    gen_tcp:close(Socket),
+
     {ok, {{_, 200, _}, _, Body}} = httpc:request(get, Request, [], []),
     %% bad handler with the closed socket should get replaced by
     %% the new one, so children count should stay the same
@@ -1233,6 +1348,53 @@ stream_fun_server_close(Config) when is_list(Config) ->
             ct:fail(did_not_receive_close)
     end. 
 
+%%--------------------------------------------------------------------
+server_closing_connection_on_first_response() ->
+    [{doc, "Client receives \"Connection: close\" on first response."
+      "A client that receives a \"close\" connection option MUST cease sending"
+      "requests on that connection and close the connection after reading"
+      "the response message containing the \"close\""}].
+server_closing_connection_on_first_response(Config) when is_list(Config) ->
+    ReqSrvSendOctFun =
+        fun(V, U, S) ->
+                {ok, {{V, S, _}, Headers0, []}} =
+                    httpc:request(get, {U, []}, [{version, V}], []),
+                {_, SendOctStr} =
+                    proplists:lookup("x-socket-stat-send-oct", Headers0),
+                list_to_integer(SendOctStr)
+        end,
+    V = "HTTP/1.1",
+    Url0 = url(group_name(Config), "/http_1_1_send_oct.html", Config),
+    Url1 = url(group_name(Config), "/http_1_1_send_oct_and_connection_close.html", Config),
+    %% Test case assumes at most one reusable past session.
+    _ = ReqSrvSendOctFun(V, Url1, 204),
+    0 = ReqSrvSendOctFun(V, Url0, 204),
+    ok.
+
+%%--------------------------------------------------------------------
+server_closing_connection_on_second_response() ->
+    [{doc, "Client receives \"Connection: close\" on second response."
+      "A client that receives a \"close\" connection option MUST cease sending"
+      "requests on that connection and close the connection after reading"
+      "the response message containing the \"close\""}].
+server_closing_connection_on_second_response(Config) when is_list(Config) ->
+    ReqSrvSendOctFun =
+        fun(V, U, S) ->
+                {ok, {{V, S, _}, Headers0, []}} =
+                    httpc:request(get, {U, []}, [{version, V}], []),
+                {_, SendOctStr} =
+                    proplists:lookup("x-socket-stat-send-oct", Headers0),
+                list_to_integer(SendOctStr)
+        end,
+    V = "HTTP/1.1",
+    Url0 = url(group_name(Config), "/http_1_1_send_oct.html", Config),
+    Url1 = url(group_name(Config), "/http_1_1_send_oct_and_connection_close.html", Config),
+    %% Test case assumes no reusable past sessions.
+    SendOct0 = 0 = ReqSrvSendOctFun(V, Url0, 204),
+    case ReqSrvSendOctFun(V, Url1, 204) of SendOct1 when SendOct1 > SendOct0 -> ok end,
+    0 = ReqSrvSendOctFun(V, Url0, 204),
+    ok.
+
 %%--------------------------------------------------------------------
 slow_connection() ->
     [{doc, "Test that a request on a slow keep-alive connection won't crash the httpc_manager"}].
@@ -1264,6 +1426,26 @@ unix_domain_socket(Config) when is_list(Config) ->
     {ok, {{_,200,_}, [_ | _], _}}
         = httpc:request(get, {URL, []}, [], []).
 
+%%-------------------------------------------------------------------------
+delete_no_body(doc) ->
+    ["Test that a DELETE request without Body does not send a Content-Type header - Solves ERL-536"];
+delete_no_body(Config) when is_list(Config) ->
+    URL = url(group_name(Config), "/delete_no_body.html", Config),
+    %% Simulated server replies 500 if 'Content-Type' header is present
+    {ok, {{_,200,_}, _, _}} =
+        httpc:request(delete, {URL, []}, [], []),
+    {ok, {{_,500,_}, _, _}} =
+        httpc:request(delete, {URL, [], "text/plain", "TEST"}, [], []).
+
+%%--------------------------------------------------------------------
+request_options() ->
+    [{doc, "Test http get request with socket options against local server (IPv6)"}].
+request_options(Config) when is_list(Config) ->
+    Request  = {url(group_name(Config), "/dummy.html", Config), []},
+    {ok, {{_,200,_}, [_ | _], _ = [_ | _]}} = httpc:request(get, Request, [],
+                                                            [{socket_opts,[{ipfamily, inet6}]}]),
+    {error,{failed_connect,_ }} = httpc:request(get, Request, [], []).
+
 
 
 %%--------------------------------------------------------------------
@@ -1353,6 +1535,9 @@ url(http, End, Config) ->
     Port = proplists:get_value(port, Config),
     {ok,Host} = inet:gethostname(),
     ?URL_START ++ Host ++ ":" ++ integer_to_list(Port) ++ End;
+url(http_ipv6, End, Config) ->
+    Port = proplists:get_value(port, Config),
+    ?URL_START ++ "[::1]" ++ ":" ++ integer_to_list(Port) ++ End;
 url(https, End, Config) ->
     Port = proplists:get_value(port, Config),
     {ok,Host} = inet:gethostname(),
@@ -1397,7 +1582,11 @@ server_start(http_unix_socket, Config) ->
     {_Pid, Port} = http_test_lib:dummy_server(unix_socket, Inet, [{content_cb, ?MODULE},
                                                                   {unix_socket, Socket}]),
     Port;
-
+server_start(http_ipv6, HttpdConfig) ->
+    {ok, Pid} = inets:start(httpd, HttpdConfig),
+    Serv = inets:services_info(),
+    {value, {_, _, Info}} = lists:keysearch(Pid, 2, Serv),
+    proplists:get_value(port, Info);
 server_start(_, HttpdConfig) ->
     {ok, Pid} = inets:start(httpd, HttpdConfig),
     Serv = inets:services_info(),
@@ -1416,6 +1605,17 @@ server_config(http, Config) ->
      {mime_type, "text/plain"},
      {script_alias, {"/cgi-bin/", filename:join(ServerRoot, "cgi-bin") ++ "/"}}
     ];
+server_config(http_ipv6, Config) ->
+    ServerRoot = proplists:get_value(server_root, Config),
+    [{port, 0},
+     {server_name,"httpc_test"},
+     {server_root, ServerRoot},
+     {document_root, proplists:get_value(doc_root, Config)},
+     {bind_address, {0,0,0,0,0,0,0,1}},
+     {ipfamily, inet6},
+     {mime_type, "text/plain"},
+     {script_alias, {"/cgi-bin/", filename:join(ServerRoot, "cgi-bin") ++ "/"}}
+    ];
 server_config(http_internal, Config) ->
     ServerRoot = proplists:get_value(server_root, Config),
     [{port, 0},
@@ -1449,10 +1649,8 @@ start_apps(_) ->
     ok.
 
 ssl_config(Config) ->
-    DataDir = proplists:get_value(data_dir, Config),
-    [{certfile, filename:join(DataDir, "ssl_server_cert.pem")},
-     {verify, verify_none}
-    ].
+    SSLConf = proplists:get_value(ssl_conf, Config),
+    proplists:get_value(server_config, SSLConf).
 
 setup_server_dirs(ServerRoot, DocRoot, DataDir) ->   
     CgiDir =  filename:join(ServerRoot, "cgi-bin"),
@@ -1770,6 +1968,13 @@ auth_header([{"authorization", Value} | _]) ->
 auth_header([_ | Tail]) ->
     auth_header(Tail).
 
+content_type_header([]) ->
+    not_found;
+content_type_header([{"content-type", Value}|_]) ->
+    {ok, string:strip(Value)};
+content_type_header([_|T]) ->
+    content_type_header(T).
+
 handle_auth("Basic " ++ UserInfo, Challange, DefaultResponse) ->
     case string:tokens(base64:decode_to_string(UserInfo), ":") of
 	["alladin", "sesame"] = Auth ->
@@ -1794,6 +1999,15 @@ content_length(["content-length:" ++ Value | _]) ->
 content_length([_Head | Tail]) ->
    content_length(Tail).
 
+handle_uri("GET","/dummy.html?foo=bar",_,_,_,_) ->
+    "HTTP/1.0 200 OK\r\n\r\nTEST";
+
+handle_uri("GET","/space%20.html",_,_,_,_) ->
+    Body = "<HTML><BODY>foobar</BODY></HTML>",
+    "HTTP/1.1 200 OK\r\n" ++
+        "Content-Length:" ++ integer_to_list(length(Body)) ++ "\r\n\r\n" ++
+        Body;
+
 handle_uri(_,"/just_close.html",_,_,_,_) ->
 		close;
 handle_uri(_,"/no_content.html",_,_,_,_) ->
@@ -1859,6 +2073,23 @@ handle_uri(_,"/301.html",Port,_,Socket,_) ->
 	"Content-Length:" ++ integer_to_list(length(Body))
 	++ "\r\n\r\n" ++ Body;
 
+
+handle_uri("HEAD","/301_rel_uri.html",_,_,_,_) ->
+    NewUri = "/dummy.html",
+    "HTTP/1.1 301 Moved Permanently\r\n" ++
+	"Location:" ++ NewUri ++  "\r\n" ++
+	"Content-Length:0\r\n\r\n";
+
+handle_uri(_,"/301_rel_uri.html",_,_,_,_) ->
+    NewUri = "/dummy.html",
+  Body = "<HTML><BODY><a href=" ++ NewUri ++
+	">New place</a></BODY></HTML>",
+    "HTTP/1.1 301 Moved Permanently\r\n" ++
+	"Location:" ++ NewUri ++  "\r\n" ++
+	"Content-Length:" ++ integer_to_list(length(Body))
+	++ "\r\n\r\n" ++ Body;
+
+
 handle_uri("HEAD","/302.html",Port,_,Socket,_) ->
     NewUri = url_start(Socket) ++
 	integer_to_list(Port) ++ "/dummy.html",
@@ -2165,10 +2396,40 @@ handle_uri("GET","/v1/kv/foo",_,_,_,_) ->
         "Content-Length: 24\r\n" ++
         "Content-Type: application/json\r\n\r\n" ++
         "[{\"Value\": \"aGVsbG8=\"}]\n";
-
+handle_uri(_,"/http_1_1_send_oct.html",_,_,Socket,_) ->
+    "HTTP/1.1 204 No Content\r\n" ++
+        "X-Socket-Stat-Send-Oct: " ++ integer_to_list(get_stat(Socket, send_oct)) ++ "\r\n" ++
+        "\r\n";
+handle_uri(_,"/http_1_1_send_oct_and_connection_close.html",_,_,Socket,_) ->
+    "HTTP/1.1 204 No Content\r\n" ++
+        "X-Socket-Stat-Send-Oct: " ++ integer_to_list(get_stat(Socket, send_oct)) ++ "\r\n" ++
+        "Connection: close\r\n" ++
+        "\r\n";
+handle_uri(_,"/delete_no_body.html", _,Headers,_, DefaultResponse) ->
+    Error = "HTTP/1.1 500 Internal Server Error\r\n" ++
+	"Content-Length:0\r\n\r\n",
+    case content_type_header(Headers) of
+	{ok, _} ->
+	    Error;
+	not_found ->
+	    DefaultResponse
+    end;
 handle_uri(_,_,_,_,_,DefaultResponse) ->
     DefaultResponse.
 
+get_stat(S, Opt) ->
+    case getstat(S, [Opt]) of
+        {ok, [{Opt, V}]} when is_integer(V) ->
+            V;
+        {error, _} = E ->
+            E
+    end.
+
+getstat(#sslsocket{} = S, Opts) ->
+    ssl:getstat(S, Opts);
+getstat(S, Opts) ->
+    inet:getstat(S, Opts).
+
 url_start(#sslsocket{}) ->
     {ok,Host} = inet:gethostname(),
     ?TLS_URL_START ++ Host ++ ":";
