@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2003-2016. All Rights Reserved.
+%% Copyright Ericsson AB 2003-2018. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -25,7 +25,7 @@
 %% Test functions
 -export([all/0, suite/0,groups/0,init_per_group/2,end_per_group/2,
 	 start_stop/1,load_file/1,not_found_items/1,
-	 non_existing/1,not_a_crashdump/1,old_crashdump/1]).
+	 non_existing/1,not_a_crashdump/1,old_crashdump/1,new_crashdump/1]).
 -export([init_per_suite/1, end_per_suite/1]).
 -export([init_per_testcase/2, end_per_testcase/2]).
 
@@ -83,6 +83,7 @@ all() ->
      non_existing,
      not_a_crashdump,
      old_crashdump,
+     new_crashdump,
      load_file,
      not_found_items
     ].
@@ -212,6 +213,25 @@ not_a_crashdump(Config) when is_list(Config) ->
 
     ok = crashdump_viewer:stop().
 
+%% Try to load a file with newer version than this crashdump viewer can handle
+new_crashdump(Config) ->
+    Dump = hd(?config(dumps,Config)),
+    ok = start_backend(Dump),
+    {ok,{MaxVsn,CurrentVsn}} = crashdump_viewer:get_dump_versions(),
+    if MaxVsn =/= CurrentVsn ->
+            ct:fail("Current dump version is not equal to cdv's max version");
+       true ->
+            ok
+    end,
+    ok = crashdump_viewer:stop(),
+    NewerVsn = lists:join($.,[integer_to_list(X+1) || X <- MaxVsn]),
+    PrivDir = ?config(priv_dir,Config),
+    NewDump = filename:join(PrivDir,"new_erl_crash.dump"),
+    ok = file:write_file(NewDump,"=erl_crash_dump:"++NewerVsn++"\n"),
+    {error, Reason} = start_backend(NewDump),
+    "This Crashdump Viewer is too old" ++_ = Reason,
+    ok = crashdump_viewer:stop().
+
 %% Load files into the tool and view all pages
 load_file(Config) when is_list(Config) ->
     case ?t:is_debug() of
@@ -328,7 +348,7 @@ browse_file(File) ->
 
     io:format("  info read",[]),
 
-    lookat_all_pids(Procs),
+    lookat_all_pids(Procs,is_truncated(File),incomplete_allowed(File)),
     io:format("  pids ok",[]),
     lookat_all_ports(Ports),
     io:format("  ports ok",[]),
@@ -339,6 +359,21 @@ browse_file(File) ->
 
     Procs. % used as second arg to special/2
 
+is_truncated(File) ->
+    case filename:extension(File) of
+        ".trunc"++_ ->
+            true;
+        _ ->
+            false
+    end.
+
+incomplete_allowed(File) ->
+    %% Incomplete heap is allowed for native libs, since some literals
+    %% are not dumped - and for pre OTP-20 (really pre 20.2) releases,
+    %% since literals were not dumped at all then.
+    Rel = get_rel_from_dump_name(File),
+    Rel < 20 orelse test_server:is_native(lists).
+
 special(File,Procs) ->
     case filename:extension(File) of
 	".full_dist" ->
@@ -364,6 +399,14 @@ special(File,Procs) ->
 		crashdump_viewer:expand_binary({SOffset,SSize,SPos}),
 	    io:format("  expand binary ok",[]),
 
+            Binaries = crashdump_helper:create_binaries(),
+            verify_binaries(Binaries, proplists:get_value(bins,Dict)),
+	    io:format("  binaries ok",[]),
+
+            SubBinaries = crashdump_helper:create_sub_binaries(Binaries),
+            verify_binaries(SubBinaries, proplists:get_value(sub_bins,Dict)),
+	    io:format("  sub binaries ok",[]),
+
 	    #proc{last_calls=LastCalls} = ProcDetails,
             true = length(LastCalls) =< 4,
 
@@ -416,13 +459,36 @@ special(File,Procs) ->
 			old_attrib=undefined,
 			old_comp_info=undefined}=Mod2,
 	    ok;
+        ".trunc_mod" ->
+            ModName = atom_to_list(?helper_mod),
+            {ok,Mod=#loaded_mod{},[TW]} =
+                crashdump_viewer:loaded_mod_details(ModName),
+            "WARNING: The crash dump is truncated here."++_ = TW,
+            #loaded_mod{current_attrib=CA,current_comp_info=CCI,
+                        old_attrib=OA,old_comp_info=OCI} = Mod,
+            case lists:all(fun(undefined) ->
+                                   true;
+                              (S) when is_list(S) ->
+                                   io_lib:printable_unicode_list(lists:flatten(S));
+                              (_) -> false
+                           end,
+                           [CA,CCI,OA,OCI]) of
+                true ->
+                    ok;
+                false ->
+                    ct:fail({should_be_printable_strings_or_undefined,
+                             {CA,CCI,OA,OCI}})
+            end,
+            ok;
 	".trunc_bin1" ->
             %% This is 'full_dist' truncated after the first
             %% "=binary:"
             %% i.e. no binary exist in the dump
 	    [#proc{pid=Pid0}|_Rest] = lists:keysort(#proc.name,Procs),
 	    Pid = pid_to_list(Pid0),
-	    {ok,ProcDetails=#proc{},[]} = crashdump_viewer:proc_details(Pid),
+            %%WarnIncompleteHeap = ["WARNING: This process has an incomplete heap. Some information might be missing."],
+	    {ok,ProcDetails=#proc{},[]} =
+                crashdump_viewer:proc_details(Pid),
 	    io:format("  process details ok",[]),
 
 	    #proc{dict=Dict} = ProcDetails,
@@ -512,21 +578,57 @@ special(File,Procs) ->
                 crashdump_viewer:ets_tables(Pid),
             io:format("  unicode table name ok",[]),
 
+            ok;
+        ".maps" ->
+	    %% I registered a process as aaaaaaaa_maps in the map dump
+	    %% to make sure it will be the first in the list when sorted
+	    %% on names.
+	    [#proc{pid=Pid0,name=Name}|_Rest] = lists:keysort(#proc.name,Procs),
+            "aaaaaaaa_maps" = Name,
+	    Pid = pid_to_list(Pid0),
+	    {ok,ProcDetails=#proc{},[]} = crashdump_viewer:proc_details(Pid),
+	    io:format("  process details ok",[]),
+
+	    #proc{dict=Dict} = ProcDetails,
+            %% io:format("~p\n", [Dict]),
+            Maps = crashdump_helper:create_maps(),
+            Maps = proplists:get_value(maps,Dict),
+            io:format("  maps ok",[]),
             ok;
 	_ ->
 	    ok
     end,
     ok.
 
+verify_binaries([H|T1], [H|T2]) ->
+    %% Heap binary.
+    verify_binaries(T1, T2);
+verify_binaries([Bin|T1], [['#CDVBin',Offset,Size,Pos]|T2]) ->
+    %% Refc binary.
+    {ok,<<Bin:Size/binary>>} = crashdump_viewer:expand_binary({Offset,Size,Pos}),
+    verify_binaries(T1, T2);
+verify_binaries([], []) ->
+    ok.
 
-lookat_all_pids([]) ->
+lookat_all_pids([],_,_) ->
     ok;
-lookat_all_pids([#proc{pid=Pid0}|Procs]) ->
+lookat_all_pids([#proc{pid=Pid0}|Procs],TruncAllowed,IncompAllowed) ->
     Pid = pid_to_list(Pid0),
-    {ok,_ProcDetails=#proc{},_ProcTW} = crashdump_viewer:proc_details(Pid),
-    {ok,_Ets,_EtsTW} = crashdump_viewer:ets_tables(Pid),
-    {ok,_Timers,_TimersTW} = crashdump_viewer:timers(Pid),
-    lookat_all_pids(Procs).
+    {ok,_ProcDetails=#proc{},ProcTW} = crashdump_viewer:proc_details(Pid),
+    {ok,_Ets,EtsTW} = crashdump_viewer:ets_tables(Pid),
+    {ok,_Timers,TimersTW} = crashdump_viewer:timers(Pid),
+    case {ProcTW,EtsTW,TimersTW} of
+        {[],[],[]} ->
+            ok;
+        {["WARNING: This process has an incomplete heap."++_],[],[]}
+          when IncompAllowed ->
+            ok;  % native libs, literals might not be included in dump
+        _ when TruncAllowed ->
+            ok; % truncated dump
+        TWs ->
+            ct:fail({unexpected_warning,TWs})
+    end,
+    lookat_all_pids(Procs,TruncAllowed,IncompAllowed).
 
 lookat_all_ports([]) ->
     ok;
@@ -572,23 +674,37 @@ do_create_dumps(DataDir,Rel) ->
 	end,
     case Rel of
 	current ->
-	    CD3 = dump_with_args(DataDir,Rel,"instr","+Mim true"),
+	    CD3 = dump_with_args(DataDir,Rel,"instr","+Muatags true"),
 	    CD4 = dump_with_strange_module_name(DataDir,Rel,"strangemodname"),
-	    Tmp = dump_with_args(DataDir,Rel,"trunc_bytes",""),
-            {ok,#file_info{size=Max}} = file:read_file_info(Tmp),
-            ok = file:delete(Tmp),
-            Bytes = max(15,rand:uniform(Max)),
-            CD5 = dump_with_args(DataDir,Rel,"trunc_bytes",
-                                 "-env ERL_CRASH_DUMP_BYTES " ++
-                                     integer_to_list(Bytes)),
+            CD5 = dump_with_size_limit_reached(DataDir,Rel,"trunc_bytes"),
             CD6 = dump_with_unicode_atoms(DataDir,Rel,"unicode"),
-            TruncatedDumps = truncate_dump(CD1),
-	    {[CD1,CD2,CD3,CD4,CD5,CD6|TruncatedDumps], DosDump};
+            CD7 = dump_with_maps(DataDir,Rel,"maps"),
+            TruncDumpMod = truncate_dump_mod(CD1),
+            TruncatedDumpsBinary = truncate_dump_binary(CD1),
+	    {[CD1,CD2,CD3,CD4,CD5,CD6,CD7,TruncDumpMod|TruncatedDumpsBinary],
+             DosDump};
 	_ ->
 	    {[CD1,CD2], DosDump}
     end.
 
-truncate_dump(File) ->
+truncate_dump_mod(File) ->
+    {ok,Bin} = file:read_file(File),
+    ModNameBin = atom_to_binary(?helper_mod,latin1),
+    NewLine = case os:type() of
+                  {win32,_} -> <<"\r\n">>;
+                  _ -> <<"\n">>
+              end,
+    RE = <<NewLine/binary,"=mod:",ModNameBin/binary,
+           NewLine/binary,"Current size: [0-9]*",
+           NewLine/binary,"Current attributes: ...">>,
+    {match,[{Pos,Len}]} = re:run(Bin,RE),
+    Size = Pos + Len,
+    <<Truncated:Size/binary,_/binary>> = Bin,
+    DumpName = filename:rootname(File) ++ ".trunc_mod",
+    file:write_file(DumpName,Truncated),
+    DumpName.
+
+truncate_dump_binary(File) ->
     {ok,Bin} = file:read_file(File),
     BinTag = <<"\n=binary:">>,
     Colon = <<":">>,
@@ -596,7 +712,10 @@ truncate_dump(File) ->
                   {win32,_} -> <<"\r\n">>;
                   _ -> <<"\n">>
               end,
-    [StartBin,AfterTag] = binary:split(Bin,BinTag),
+    %% Split after "our binary" created by crashdump_helper
+    %% (it may not be the first binary).
+    RE = <<"\n=binary:(?=[0-9A-Z]+",NewLine/binary,"FF:AQID)">>,
+    [StartBin,AfterTag] = re:split(Bin,RE,[{parts,2}]),
     [AddrAndSize,BinaryAndRest] = binary:split(AfterTag,Colon),
     [Binary,_Rest] = binary:split(BinaryAndRest,NewLine),
     TruncSize = byte_size(Binary) - 2,
@@ -689,6 +808,28 @@ dump_with_strange_module_name(DataDir,Rel,DumpName) ->
     ?t:stop_node(n1),
     CD.
 
+dump_with_size_limit_reached(DataDir,Rel,DumpName) ->
+    Tmp = dump_with_args(DataDir,Rel,DumpName,""),
+    {ok,#file_info{size=Max}} = file:read_file_info(Tmp),
+    ok = file:delete(Tmp),
+    dump_with_size_limit_reached(DataDir,Rel,DumpName,Max).
+
+dump_with_size_limit_reached(DataDir,Rel,DumpName,Max) ->
+    Bytes = max(15,rand:uniform(Max)),
+    CD = dump_with_args(DataDir,Rel,DumpName,
+                        "-env ERL_CRASH_DUMP_BYTES " ++
+                            integer_to_list(Bytes)),
+    {ok,#file_info{size=Size}} = file:read_file_info(CD),
+    if Size =< Bytes ->
+            %% This means that the dump was actually smaller than the
+            %% randomly selected truncation size, so we'll just do it
+            %% again with a smaller number
+            ok = file:delete(CD),
+            dump_with_size_limit_reached(DataDir,Rel,DumpName,Size-3);
+       true ->
+            CD
+    end.
+
 dump_with_unicode_atoms(DataDir,Rel,DumpName) ->
     Opt = rel_opt(Rel),
     Pz = "-pz \"" ++ filename:dirname(code:which(?MODULE)) ++ "\"",
@@ -699,6 +840,16 @@ dump_with_unicode_atoms(DataDir,Rel,DumpName) ->
     ?t:stop_node(n1),
     CD.
 
+dump_with_maps(DataDir,Rel,DumpName) ->
+    Opt = rel_opt(Rel),
+    Pz = "-pz \"" ++ filename:dirname(code:which(?MODULE)) ++ "\"",
+    PzOpt = [{args,Pz}],
+    {ok,N1} = ?t:start_node(n1,peer,Opt ++ PzOpt),
+    {ok,_Pid} = rpc:call(N1,crashdump_helper,dump_maps,[]),
+    CD = dump(N1,DataDir,Rel,DumpName),
+    ?t:stop_node(n1),
+    CD.
+
 dump(Node,DataDir,Rel,DumpName) ->
     Crashdump = filename:join(DataDir, dump_prefix(Rel)++DumpName),
     rpc:call(Node,os,putenv,["ERL_CRASH_DUMP",Crashdump]),
@@ -751,6 +902,11 @@ dump_prefix(current) ->
 dump_prefix(Rel) ->
     lists:concat(["r",Rel,"_dump."]).
 
+get_rel_from_dump_name(File) ->
+    Name = filename:basename(File),
+    ["r"++Rel|_] = string:split(Name,"_"),
+    list_to_integer(Rel).
+
 compat_rel(current) ->
     "";
 compat_rel(Rel) ->
