@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2008-2017. All Rights Reserved.
+%% Copyright Ericsson AB 2008-2018. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -41,8 +41,7 @@
 %%--------------------------------------------------------------------
 
 suite() ->
-    [%%{ct_hooks,[ts_install_cth]},
-     {timetrap,{seconds,40}}].
+    [{timetrap,{seconds,60}}].
 
 all() ->
 %%    [check_docker_present] ++
